import time
from builtins import super

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal

from .expectation_maximization_registration import expectation_maximization_registration


class rigid_registration(expectation_maximization_registration):
    """
    Rigid registration class.

    Parameters
    ----------
    R : array_like, optional
        The rotation matrix.

    t : array_like, optional
        The translation vector.

    s : array_like, optional
        The scale factor.

    scale : bool, optional
        Whether to lean scale for the point cloud.
    """

    def __init__(self, R=None, t=None, s=None, scale=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        tic = time.time()
        if self.D != 2 and self.D != 3:
            message = (
                "Rigid registration only supports 2D or 3D point clouds. Instead got {}.".format(
                    self.D
                )
            )
            raise ValueError(message)
        if s == 0:
            raise ValueError("A zero scale factor is not supported.")
        self.R = np.eye(self.D) if R is None else R
        self.t = np.atleast_2d(np.zeros((1, self.D))) if t is None else t
        self.s = 1.0 if s is None else s
        self.muX = None
        self.muY = None
        self.myU = None
        self.S = None
        self.C = None
        self.scale = scale

        toc = time.time()
        self.time_to_initialize_registration = toc - tic

    def update_transform(self):
        """
        Update the transform parameters.
        """
        # Replaced the PyCPD muX and muY because couldnt calcl myY the same (no self.P)
        # and wanted to use the same method for both.
        self.muX = np.divide(np.matmul(self.X.T, self.Pt1), self.Np)
        self.muY = np.divide(
            np.matmul(self.Y.T, self.P1), self.Np
        )  # changed previous code from PyCPD because we no longer store self.P
        A = np.matmul(self.PX.T, self.Y) - self.Np * np.matmul(
            self.muX[:, None], self.muY[:, None].T
        )  # A= X'P'*Y-X'P'1*1'P'Y/Np;

        U, self.S, V = np.linalg.svd(
            A, full_matrices=True
        )  # matlab and numpy return V in different orientations.
        self.S = np.diag(self.S)
        self.C = np.eye(self.D)
        # The following line means that we allow reflections. Can make this an option. See line 96:
        # https://github.com/markeroon/matlab-computer-vision-routines/blob/master/third_party/CoherentPointDrift/core/Rigid/cpd_rigid.m#L96
        self.C[self.D - 1, self.D - 1] = np.linalg.det(np.dot(U, V))

        self.R = np.transpose(np.dot(np.dot(U, self.C), V))
        if self.scale is True:
            self.s = np.trace(np.matmul(self.S, self.C)) / (
                np.sum(np.square(self.Y) * self.P1[:, None])
                - self.Np * (np.matmul(self.muY.T, self.muY))
            )
        elif self.scale is False:
            self.s = 1
        self.t = self.muX - self.s * np.dot(self.R.T, self.muY)

        # Use E (matlab function cpd_P) or L (matlab code for registration) to calculate the "error".
        # This is instead of using sigma which was the approach used by PyCPD. Doing this to ensure consistency between
        # This and the original implementation.
        # self.err = abs((self.E - self.E_old) / self.E)
        self.err = abs(
            self.E - self.E_old
        )  # The above is a "normalized" alternative that was used in the Matlab code.

    def transform_point_cloud(self, Y=None):
        """
        Transform a point cloud.

        Parameters
        ----------
        Y : array_like, optional
            The point cloud to transform.

        """
        if Y is None:
            self.TY = self.s * np.dot(self.Y, self.R) + self.t
            return
        else:
            return self.s * np.dot(Y, self.R) + self.t

    def update_variance(self):
        """
        Update the variance.
        """
        self.sigma2_prev = self.sigma2
        # A, B, C were just implemented to make the code reasier to read/interpret.
        # The code for self.sigma2 matches what is written in the matlab rigid example.
        A = np.sum(np.square(self.X) * self.Pt1[:, None])
        B = self.Np * np.matmul(self.muX[:, None].T, self.muX[:, None])

        if self.scale is True:
            C = self.s * np.trace(np.matmul(self.S, self.C))
            self.sigma2 = abs(A - B - C) / (self.Np * self.D)
        elif self.scale is False:
            C = np.sum(np.square(self.Y) * self.P1[:, None])
            D = self.Np * np.matmul(self.muY[:, None].T, self.muY[:, None])
            E = 2.0 * np.trace(np.matmul(self.S, self.C))
            self.sigma2 = abs(A - B + C - D - E) / (self.Np * self.D)
        # Still retaining the replacement/update of sigma2 if <=0 that was in PyCPD. This might not be be
        # necessary any more.
        if self.sigma2 <= 0:
            self.sigma2 = self.tolerance / 10

    def get_registration_parameters(self):
        """
        Get the registration parameters.

        Returns
        -------
        self.s : float
            The scale factor.

        self.R : array_like
            The rotation matrix.

        self.t : array_like
            The translation vector.
        """
        return self.s, self.R, self.t



import time
from builtins import super

import numpy as np

from .expectation_maximization_registration import expectation_maximization_registration


class affine_registration(expectation_maximization_registration):
    """ """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        tic = time.time()
        self.B = np.eye(self.D)
        self.t = np.zeros(self.D)
        self.B1 = None
        self.B2 = None

        toc = time.time()
        self.time_to_initialize_registration = toc - tic

    def update_transform(self):
        """
        Update the transform parameters.
        """
        self.muX = np.divide(np.matmul(self.X.T, self.Pt1), self.Np)
        self.muY = np.divide(
            np.matmul(self.Y.T, self.P1), self.Np
        )  # changed previous code from PyCPD because we no longer store self.P

        self.B1 = np.matmul(self.PX.T, self.Y) - self.Np * np.matmul(
            self.muX[:, None], self.muY[:, None].T
        )
        self.B2 = np.matmul((self.Y * self.P1[:, None]).T, self.Y) - self.Np * np.matmul(
            self.muY[:, None], self.muY[:, None].T
        )
        self.B = np.linalg.solve(self.B2.T, self.B1.T).T

        self.t = np.squeeze(self.muX[:, None] - np.matmul(self.B, self.muY[:, None]))

        self.err = abs(self.E - self.E_old)

    def transform_point_cloud(self, Y=None):
        """
        Transform a point cloud.

        Parameters
        ----------
        Y : array_like, optional
            The point cloud to transform.
        """
        if Y is None:
            self.TY = np.dot(self.Y, self.B.T) + self.t
            return
        else:
            return np.dot(Y, self.B.T) + self.t

    def update_variance(self):
        """
        Update the variance.
        """
        self.sigma2_prev = self.sigma2
        A = np.sum(np.square(self.X) * self.Pt1[:, None])
        B = self.Np * np.matmul(self.muX[:, None].T, self.muX[:, None])
        C = np.trace(np.matmul(self.B1, self.B.T))
        self.sigma2 = abs(A - B - C) / (self.Np * self.D)

        if self.sigma2 <= 0:
            self.sigma2 = self.tolerance / 10

    def get_registration_parameters(self):
        """
        Get the registration parameters.

        Returns
        -------
        self.B : array_like
            The affine transform matrix.
        self.t : array_like
            The translation vector.
        """
        return self.B, self.t



"""
Numpy + Cython Implementation of the Coherent Point Drift
([CPD](https://arxiv.org/abs/0905.2635/)) algorithm by 
Myronenko and Song. This implementation aims to speed up the 
PyCPD implementation of CPD by using Cython. It provides 
three registration methods for point clouds: 

1. Scale and rigid registration
2. Affine registration
3. Gaussian regularized non-rigid registration

Licensed under an MIT License (c) Anthony Gatti.
Distributed here: https://github.com/gattia/cycpd
"""

from .affine_registration import affine_registration
from .deformable_registration import deformable_registration, gaussian_kernel
from .rigid_registration import rigid_registration



import time

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal

import cython_functions as cy


class expectation_maximization_registration(object):
    """
    Expectation Maximization class

    Parameters
    ----------
    X : numpy.ndarray
        The target point cloud.

    Y : numpy.ndarray
        The source point cloud.

    max_iterations : int
        The maximum number of iterations to perform.

    tolerance : float
        The tolerance for the error.

    w : float
        Contribution of the uniform distribution to account for outliers.
        Valid values span 0 (inclusive) and 1 (exclusive).

    verbose : bool
        If True, print verbose statements to commandline.

    print_reg_params : bool
        If True, print out the registration parameters.



    """

    def __init__(
        self,
        X,
        Y,
        max_iterations=100,
        tolerance=1e-5,
        w=0.0,
        verbose=True,
        print_reg_params=True,
        *args,
        **kwargs
    ):
        if type(X) is not np.ndarray or X.ndim != 2:
            raise ValueError("The target point cloud (X) must be at a 2D numpy array.")
        if type(Y) is not np.ndarray or Y.ndim != 2:
            raise ValueError("The source point cloud (Y) must be a 2D numpy array.")
        if X.shape[1] != Y.shape[1]:
            raise ValueError("Both point clouds need to have the same number of dimensions.")

        self.expectation_times = []
        self.maximization_times = []
        self.update_transform_times = []
        self.transform_times = []
        self.update_variance_times = []
        self.time_to_initialize_registration = 0
        self.time_to_initiate_registration = 0

        self.X = X
        self.Y = Y
        self.TY = None
        self.N = self.X.shape[0]
        self.D = self.X.shape[1]
        self.M = self.Y.shape[0]
        self.tolerance = tolerance
        self.w = w
        self.max_iterations = max_iterations
        self.iteration = 0
        self.err = self.tolerance + 1
        self.P = np.zeros((self.M, self.N))
        self.Pt1 = np.zeros((self.N,))
        self.P1 = np.zeros((self.M,))
        self.PX = np.zeros((self.M, self.D))
        self.Np = 0.0
        self.E = 0.0
        self.E_old = 0.0
        self.sigma2 = 0.0
        self.sigma2_prev = 0.0
        self.verbose = verbose
        self.print_reg_params = print_reg_params

    def register(self, callback=lambda **kwargs: None):
        """
        Register the point clouds.

        Parameters
        ----------
        callback : function
            A callback function that is called after each iteration.
            This is used to print figures during the registration process.

            The callback function should take the following arguments:
                - iteration : int
                    The current iteration.
                - E : float
                    The registration error.
                - X : numpy.ndarray
                    The original point cloud.
                - Y : numpy.ndarray
                    The transformed point cloud.
        """
        tic = time.time()
        self.transform_point_cloud()
        self.sigma2 = cy.initialize_sigma2(self.X, self.Y)
        toc = time.time()
        self.time_to_initiate_registration = toc - tic
        while (
            (self.iteration < self.max_iterations)
            and (self.err > self.tolerance)
            and (self.sigma2 > np.finfo(float).eps)
        ):
            self.iterate()
            if callable(callback):
                kwargs = {"iteration": self.iteration, "error": self.err, "X": self.X, "Y": self.TY}
                callback(**kwargs)

        if self.print_reg_params is True:
            print("=" * 72)
            print("Registration Performance Metrics")
            print("=" * 72)
            print("Time to initialize EM: {}".format(self.time_to_initiate_registration))
            print(
                "Time to initialize registration: {}".format(self.time_to_initialize_registration)
            )
            print(
                "Average Expectation Time:                {:10.4f} +/- {:5.4f}".format(
                    np.mean(self.expectation_times), np.std(self.expectation_times)
                )
            )
            print(
                "Average Maximization Time:               {:10.4f} +/- {:5.4f}".format(
                    np.mean(self.maximization_times), np.std(self.maximization_times)
                )
            )
            print("Maximization Times - Per individual step")
            print(
                "Average Update Transform Time:           {:10.4f} +/- {:5.4f}".format(
                    np.mean(self.update_transform_times), np.std(self.update_transform_times)
                )
            )
            print(
                "Average Transform Time:                  {:10.4f} +/- {:5.4f}".format(
                    np.mean(self.transform_times), np.std(self.transform_times)
                )
            )
            print(
                "Average Update Variance Time:            {:10.4f} +/- {:5.4f}".format(
                    np.mean(self.update_variance_times), np.std(self.update_variance_times)
                )
            )
            print("")
            print("Number of iterations performed:          {}".format(self.iteration))
            print("Error at time of finish:                 {}".format(self.err))

        return self.TY, self.get_registration_parameters()

    def get_registration_parameters(self):
        raise NotImplementedError("Registration parameters should be defined in child classes.")

    def iterate(self):
        """
        Perform one iteration of the Expectation Maximization algorithm.
        """
        self.E_old = self.E  # E = negative log-likelihood function
        tic = time.time()
        self.expectation()
        toc = time.time()
        self.expectation_times.append(toc - tic)

        tic = time.time()
        self.maximization()
        toc = time.time()
        self.maximization_times.append(toc - tic)

        self.iteration += 1

        if self.verbose is True:
            while type(self.sigma2) is np.ndarray:
                self.sigma2 = self.sigma2[0]

            print("Iteration:{}".format(self.iteration))
            print(
                "ML:{:10.3f}; \t"
                "ML change (error):{:10.3f}; \t"
                "Sigma^2:{:10.3f}; \t"
                "Sigma^2 change:{:10.3f}".format(
                    self.E, self.err, self.sigma2, self.sigma2_prev - self.sigma2
                )
            )
            percent_done = int(72 * self.iteration / self.max_iterations)
            print("[" + "=" * percent_done + " " * (72 - percent_done) + "]")

    def expectation(self):
        """
        Perform the Expectation step of the Expectation Maximization algorithm.
        """
        self.P1, self.Pt1, self.PX, self.Np, self.E = cy.expectation_2(
            self.X, self.TY, self.sigma2, self.M, self.N, self.D, self.w
        )

    def maximization(self):
        """
        Perform the Maximization step of the Expectation Maximization algorithm.
        """
        tic = time.time()
        self.update_transform()
        toc = time.time()
        self.update_transform_times.append(toc - tic)

        tic = time.time()
        self.transform_point_cloud()
        toc = time.time()
        self.transform_times.append(toc - tic)

        tic = time.time()
        self.update_variance()
        toc = time.time()
        self.update_variance_times.append(toc - tic)



import time
from builtins import super

import numpy as np

from .expectation_maximization_registration import expectation_maximization_registration


def gaussian_kernel(X, beta, Y=None):
    """
    Compute the Gaussian kernel matrix.

    Parameters
    ----------
    X : numpy.ndarray
        The point cloud.

    beta : float
        The bandwidth of the Gaussian kernel.

    Y : numpy.ndarray, optional
        Second point cloud to compute the kernel matrix for.
        If None, the second kernel matrix is X.

    Returns
    -------
    G : numpy.ndarray
        The Gaussian kernel matrix.
    """
    if Y is None:
        Y = X
    diff = X[:, None, :] - Y[None, :, :]
    diff = diff**2
    diff = np.sum(diff, 2)
    return np.exp(-diff / (2 * beta**2))


def lowrankQS(G, num_eig, eig_fgt=False):
    """
    Compute the low-rank approximation of the kernel matrix.

    !!!
    This function is a placeholder for implementing the fast
    gauss transform. It is not yet implemented.
    !!!

    Parameters
    ----------
    G : numpy.ndarray
        The gaussian kernel matrix.

    num_eig : int
        The number of eigenvectors to use.

    eig_fgt : bool
        If True, use the fast gauss transform to speed up registration.

    """
    # if we do not use FGT we construct affinity matrix G and find the
    # first eigenvectors/values directly

    if eig_fgt is False:
        S, Q = np.linalg.eigh(G)
        eig_indices = list(np.argsort(np.abs(S))[::-1][:num_eig])
        Q = Q[:, eig_indices]  # eigenvectors
        S = S[eig_indices]  # eigenvalues.

        return Q, S

    elif eig_fgt is True:
        raise Exception("Fast Gauss Transform Not Yet Implemented!")


class deformable_registration(expectation_maximization_registration):
    def __init__(self, alpha=2, beta=2, low_rank=True, num_eig=100, eig_fgt=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        tic = time.time()
        self.alpha = 2 if alpha is None else alpha
        self.beta = 2 if beta is None else beta
        self.num_eig = num_eig
        self.eig_fgt = eig_fgt
        self.W = np.zeros((self.M, self.D))
        self.G = gaussian_kernel(self.Y, self.beta)
        self.low_rank = low_rank

        if self.low_rank is True:
            self.Q, self.S = lowrankQS(self.G, num_eig=self.num_eig, eig_fgt=self.eig_fgt)
            self.inv_S = np.diag(1.0 / self.S)
            self.S = np.diag(self.S)
        toc = time.time()
        self.time_to_initialize_registration = toc - tic

    def update_transform(self):
        """
        Update the transform parameters.
        """
        if self.low_rank is False:
            tic = time.time()
            A = np.dot(np.diag(self.P1), self.G) + self.alpha * self.sigma2 * np.eye(self.M)
            toc = time.time()
            self.A_times.append(toc - tic)
            tic = time.time()
            B = self.PX - np.dot(np.diag(self.P1), self.Y)
            toc = time.time()
            self.B_times.append(toc - tic)
            tic = time.time()
            self.W = np.linalg.solve(A, B)
            toc = time.time()
            self.solve_times.append(toc - tic)
        else:
            dP = np.diag(self.P1)
            dPQ = np.matmul(dP, self.Q)
            F = self.PX - np.matmul(dP, self.Y)

            self.W = (
                1
                / (self.alpha * self.sigma2)
                * (
                    F
                    - np.matmul(
                        dPQ,
                        (
                            np.linalg.solve(
                                (self.alpha * self.sigma2 * self.inv_S + np.matmul(self.Q.T, dPQ)),
                                (np.matmul(self.Q.T, F)),
                            )
                        ),
                    )
                )
            )
            QtW = np.matmul(self.Q.T, self.W)
            self.E = self.E + self.alpha / 2 * np.trace(np.matmul(QtW.T, np.matmul(self.S, QtW)))
            # self.err = abs((self.E - self.E_old) / self.E)
            self.err = abs(self.E - self.E_old)
            # The absolute difference is more conservative (does more iterations) than the line above it which
            # is calculating the normalized change in the E(L). This calculation was changed to match the matlab
            # code created for low_rank matrices.

    def transform_point_cloud(self, Y=None):
        """
        Transform the point cloud.

        Parameters
        ----------
        Y : numpy.ndarray, optional
            The point cloud to transform. If None, the inpputted source
            point cloud is used.

        Returns
        -------
        Y_transformed : numpy.ndarray
            The transformed point cloud.
        """
        if self.low_rank is False:
            if Y is None:
                self.TY = self.Y + np.dot(self.G, self.W)
                return
            else:
                G = gaussian_kernel(Y, self.beta, Y=self.Y)
                return Y + np.matmul(G, self.W)
        elif self.low_rank is True:
            if Y is None:
                self.TY = self.Y + np.matmul(self.Q, np.matmul(self.S, np.matmul(self.Q.T, self.W)))
                return
            else:
                G = gaussian_kernel(Y, beta=self.beta, Y=self.Y)
                return Y + np.matmul(G, self.W)

    def update_variance(self):
        """
        Update the variance parameters.
        """
        self.sigma2_prev = self.sigma2
        A = np.sum(np.square(self.X) * self.Pt1[:, None])
        B = np.sum(np.square(self.TY) * self.P1[:, None])
        C = 2 * np.trace(np.matmul(self.PX.T, self.TY))
        self.sigma2 = abs(A + B - C) / (self.Np * self.D)

        if self.sigma2 <= 0:
            self.sigma2 = self.tolerance / 10

        # self.err = abs(self.sigma2_prev - self.sigma2)

    def get_registration_parameters(self):
        """
        Get the registration parameters.

        Returns
        -------
        self.G : numpy.ndarray
            The Gaussian kernel matrix.

        self.W : numpy.ndarray
            The transform parameters.
        """
        return self.G, self.W



from setuptools import Extension, setup

# potential way to read in version # ?
# https://blogs.nopcode.org/brainstorm/2013-05-20-pragmatic-python-versioning-via-setuptools-and-git-tags/


def readme():
    with open("README.MD") as f:
        return f.read()


try:
    from Cython.Build import cythonize

    ext_modules = cythonize(
        [
            Extension(
                name="cython_functions",
                sources=["cycpd/cython/cython_functions.pyx"],
                include_dirs=["cycpd/cython/"],
            )
        ]
    )

except ImportError:
    ext_modules = None

try:
    import numpy as np
except ImportError:
    raise Exception(
        "Numpy must be installed to build this pacakge.\n Install with `pip install .` or run `pip install -r requirements.txt` before building."
    )

setup(
    name="cycpd",
    version="0.23",
    description="Numpy + Cython Implementation of the Coherent Point Drift Algorithm",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/gattia/cycpd",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        # 'Programming Language :: Python :: 3.4',
        # 'Programming Language :: Python :: 3.5',
        # 'Programming Language :: Python :: 3.6',
        "Topic :: Scientific/Engineering",
    ],
    keywords="image processing, point cloud, registration, mesh, surface",
    author="Anthony Gatti",
    author_email="anthony@neuralseg.com",
    license="MIT",
    ext_modules=ext_modules,
    # include_dirs=[np.get_include()],
    packages=["cycpd"],
    # package_data={"cycpd.cython": ["cython_functions.pxd"]},
    setup_requires=["Cython>=0.29"],
    install_requires=["numpy"],
    zip_safe=False,
)
